==> lhcsmapi-1.1.15.tar.gz: lhcsmapi-1.1.15/lhcsmapi/dbsignal/SignalAnalysis.py <==
import numpy as np
from scipy import signal as s
import pandas as pd
from scipy.spatial.distance import euclidean, pdist, squareform
import math

class SignalAnalysis:

    @staticmethod
    def findRowsWithExactPatternInSeries(series, pattern):
        seriesWithLastIndexOfPattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        dfWithPattern = seriesWithLastIndexOfPattern.fillna(0)
        lastIndicesOfPattern = np.where(dfWithPattern)[0]  # indices with the last found element
        allIndicesOfPattern = [range(match - len(pattern) + 1, match + 1) for match in
                               lastIndicesOfPattern]  # all indices corresponding to the pattern

        data = []
        for subs in allIndicesOfPattern:
            data.append(series.iloc[subs].index)

        # flatten indices
        flattenedIndices = [val for indicesOfPattern in allIndicesOfPattern for val in indicesOfPattern]

        return (pd.DataFrame(data, columns=pattern), flattenedIndices)

    @staticmethod
    def resampleSeries(originSeries, ts=0):
        """Function resamples series by interpolating in order to get continuous sample frequency"""
        # takes sample time of first sample if not given
        if ts == 0:
            ts = (originSeries.index[1] - originSeries.index[0])  # Sample distance
        T = (originSeries.index[-1] - originSeries.index[0])  # Timeperiod
        N = int(T / ts)  # Number of samplepoints
        indexList = np.linspace(originSeries.index[0], T, N)

        # interpolate data in order to have the same sample time everywhere
        interpolatedSeries = originSeries.reindex(originSeries.index.union(indexList)).interpolate('index', limit_direction='forward') 
        # delete those points who are not in in the continous intervall
        delIndex = [x for x in originSeries.index if x not in indexList]
        interpolatedSeries = interpolatedSeries.drop(delIndex, axis=0)
        return interpolatedSeries

    @staticmethod
    def getWindowWithGivenMeanAndStd(originSeries, indexIncrement=20, mean=50, std=0.1):
        """Method returns a series within a given mean and standart deviation"""
        for j in range(len(originSeries) - indexIncrement):
            window = originSeries.iloc[j:j + indexIncrement]
            windowMean = window.mean()
            windowStd = window.std()
            isWindowFound = windowMean >= mean and windowStd <= std
            if isWindowFound:
                minIndex = originSeries.iloc[j:].idxmin() # find the time index of minimum value
                minIndex = originSeries.index.get_loc(minIndex) # find the iteration of this index
                return originSeries.iloc[j:minIndex].copy()
        print('No window found in getWindowWithGivenMeanAndStd')
        return None

        
    @staticmethod
    def applyFrequencyFilter(originSeries, sampleFrequency=0, cutOffFrequency=0, filterType = 'low'):
        """Method applies low pass filter to series"""
        #Set cuttoff frequency if not given
        if sampleFrequency == 0:
            sampleFrequency = 1/(originSeries.index[1]-originSeries.index[0]) #Sample distance
        if cutOffFrequency == 0:
            cutOffFrequency = sampleFrequency/100  # Cut-off frequency of the filter
        normalizedFrequency = cutOffFrequency / (sampleFrequency / 2) # Normalize the frequency

        b, a = s.butter(5, normalizedFrequency, filterType)
        filteredSeries = pd.DataFrame(data = s.filtfilt(b, a, originSeries.values), index = originSeries.index, columns = ['filtered'])
        return filteredSeries['filtered']
    
    @staticmethod
    def getWindowWithDeviationFromInitialValue(originSeries, initialValue =0 ,deviationAllowance=0.998):
        """Method applies low pass filter to series"""
        if initialValue == 0:
            initialValue = np.mean(s.medfilt(originSeries.values[0:19],3))
        # use filter to find start of decay
        filteredSeries = SignalAnalysis.applyFrequencyFilter(originSeries)
        # get Df with initial values
        initialSeries = filteredSeries[filteredSeries.values > initialValue*deviationAllowance]
        # Only take values after the initial phase 
        decaySeries = originSeries[originSeries.index > initialSeries.index[-1]]
        if len(decaySeries) > 0:
            return decaySeries
        else: 
            print('No window found in getWindowWithDeviationFromInitialValue')
            return None
    
    @staticmethod
    def calculateSpectrogramFromSeries(originSeries):
        """Function calculates spectrogram from DF"""
        
        originSeries = originSeries.to_frame(name=None).set_index((originSeries.index-originSeries.index[0]))[originSeries.name]  #*10**-9
        ts = (originSeries.index[1]-originSeries.index[0]) #Sample distance
        T = (originSeries.index[-1]-originSeries.index[0]) #Timeperiod
        N = int(T/ts) # Number of samplepoints
        if len(originSeries) == N: #are the samples continuous
            newSeries = originSeries
        else:
            newSeries = SignalAnalysis.resampleSeries(originSeries)
        nperseg = N // 128
        try:
            f, t, Sxx = s.spectrogram(newSeries.values, 1/ts ,window=('hann'), nperseg=nperseg, noverlap = nperseg // 2)
            return f, t, Sxx
        except Exception as e:
            print('Could not calculate spectrogram: '+ str(e)) 
            
    
    @staticmethod
    def calculateFilterBankFromSpectrogram(spectrogram):
        """Function filterbank feature from sprectrogram"""
        if spectrogram is not None:
            f, t, Sxx = spectrogram
            filterBankMaxValues = np.zeros(len(Sxx[:,1])//8)
            for i in range(len(filterBankMaxValues)):
                k = i*8           
                filterBankMaxValues[i] = np.amax(Sxx[k:k+8,:])
            return np.mean(filterBankMaxValues[1:5])
        
    @staticmethod     
    def calculateCrossCorrelation(refSeries, measSeries):
        """Function compares to signals and returns the correlation coefficient, the length of the signals can vary"""
        shift = len(measSeries.values) - len(refSeries.values)
        maxCorrelationTimestamp = 0
        if shift > 0:
            R_val = np.zeros(abs(shift));
            i_val = np.zeros(abs(shift));

            for i in range(0 ,abs(shift)):
                i_val[i] += i;
                R = np.corrcoef(measSeries.values[i:i+len(refSeries.values)], refSeries.values);
                R_val[i] = R[0][1]

            maxCorrelationTimestamp = measSeries.index[np.argmax(R_val)]
            maxCorrelation = np.max(R_val)

        elif shift < 0:
            R_val = np.zeros(abs(shift));
            i_val = np.zeros(abs(shift));

            for i in range(0 ,abs(shift)):
                i_val[i] += i;
                R = np.corrcoef(refSeries.values[i:i+len(measSeries.values)], measSeries.values);
                R_val[i] = R[0][1]

            maxCorrelationTimestamp = refSeries.index[np.argmax(R_val)]
            maxCorrelation = np.max(R_val)
        else:
            R = np.corrcoef(measSeries.values, refSeries.values);
            maxCorrelation = R[0][1]
        return maxCorrelation ,  maxCorrelationTimestamp   
    
    @staticmethod 
    def calculateSimilarityMatrix(myDF):
        simM = np.full(np.math.factorial(len(myDF.columns)-1), -1)
        myVariable = np.nonzero(np.triu(squareform(pdist(myDF.T, metric='euclidean'))).flatten())[0]
        if len(myVariable) > 0:
            simM = np.triu(squareform(pdist(myDF.T, metric='euclidean'))).flatten()
        return pd.DataFrame(data = simM[np.nonzero(simM)], index = np.nonzero(simM)[0], columns = [myDF.columns.values[0][:-2] +'_SimilarityMatrix'])

    @staticmethod
    def calculateCurve2CurveComparison(mySeriesDecay, mySeriesDecayRef, envelopeAmplitude,
                                       envelopeDecayTimeConstant=False, applyFilter=False, filterCutOffFrequency=100):
        end = np.min([len(mySeriesDecay), len(mySeriesDecayRef)])

        mySeriesDecay = mySeriesDecay.iloc[len(mySeriesDecay) - end:]
        mySeriesDecayRef = mySeriesDecayRef.iloc[len(mySeriesDecayRef) - end:]

        if applyFilter:
            mySeriesDecay = SignalAnalysis.applyFrequencyFilter(mySeriesDecay, filterCutOffFrequency,
                                                                filterType=applyFilter)
            mySeriesDecayRef = SignalAnalysis.applyFrequencyFilter(mySeriesDecayRef, filterCutOffFrequency,
                                                                   filterType=applyFilter)

        index = mySeriesDecay.index
        if envelopeDecayTimeConstant:
            envelope = np.exp(-index / envelopeDecayTimeConstant)
            substractArray = np.abs(mySeriesDecayRef.values - mySeriesDecay.values) / envelope
        else:
            substractArray = np.abs(mySeriesDecayRef.values - mySeriesDecay.values)

        outOfEnvelopeRatio = np.sum(substractArray > envelopeAmplitude) / len(substractArray)
        return outOfEnvelopeRatio

    def findTimestampFromReference(qhReferenceDischarges, magnet):
        # Look for reference curve
        if not qhReferenceDischarges[qhReferenceDischarges['magnet'] == magnet].empty:
            return qhReferenceDischarges[qhReferenceDischarges['magnet'] == magnet]['timestamp'].values[0]
        else:
            return False

    @staticmethod
    def calculateDtwDistance(s1, s2):
        DTW = {}

        for i in range(len(s1)):
            DTW[(i, -1)] = float('inf')
        for i in range(len(s2)):
            DTW[(-1, i)] = float('inf')
        DTW[(-1, -1)] = 0

        for i in range(len(s1)):
            for j in range(len(s2)):
                dist = (s1[i] - s2[j]) ** 2
                DTW[(i, j)] = dist + min(DTW[(i - 1, j)], DTW[(i, j - 1)], DTW[(i - 1, j - 1)])

        return math.sqrt(DTW[len(s1) - 1, len(s2) - 1])

==> lhcsmapi-1.1.123.tar.gz: lhcsmapi-1.1.123/lhcsmapi/dbsignal/SignalAnalysis.py <==
import numpy as np
from scipy import signal as s
import pandas as pd
from scipy.spatial.distance import euclidean, pdist, squareform
import math

class SignalAnalysis:

    @staticmethod
    def findRowsWithExactPatternInSeries(series, pattern):
        seriesWithLastIndexOfPattern = series.rolling(len(pattern)).apply(lambda x: all(np.equal(x, pattern)),
                                                                          raw=False)
        dfWithPattern = seriesWithLastIndexOfPattern.fillna(0)
        lastIndicesOfPattern = np.where(dfWithPattern)[0]  # indices with the last found element
        allIndicesOfPattern = [range(match - len(pattern) + 1, match + 1) for match in
                               lastIndicesOfPattern]  # all indices corresponding to the pattern

        data = []
        for subs in allIndicesOfPattern:
            data.append(series.iloc[subs].index)

        # flatten indices
        flattenedIndices = [val for indicesOfPattern in allIndicesOfPattern for val in indicesOfPattern]

        return (pd.DataFrame(data, columns=pattern), flattenedIndices)

    @staticmethod
    def resampleSeries(originSeries, ts=0, t0=None, t_end=None):
        """Function resamples series by interpolating in order to get continuous sample frequency"""
        # takes sample time of first sample if not given
        if ts == 0:
            ts = (originSeries.index[1] - originSeries.index[0])  # Sample distance
        if t0 is not None and t_end is not None:
            indexList = SignalAnalysis.getTimePointDistribution(ts, t0, t_end)
        else:
            indexList = SignalAnalysis.getTimePointDistribution(ts, originSeries.index[0], originSeries.index[-1])

        # interpolate data in order to have the same sample time everywhere
        interpolatedSeries = originSeries.reindex(originSeries.index.union(indexList)).interpolate('index',
                                                                                                   limit_direction='forward')
        # delete those points which are not in in the continuous interval
        delIndex = [x for x in originSeries.index if x not in indexList]
        return interpolatedSeries.drop(delIndex, axis=0)

    @staticmethod
    def getTimePointDistribution(ts, t0, t_end):
        T = (t_end - t0)  # Time period
        N = int(T / ts)  # Number of sample points
        return np.linspace(t0, T + t0, N)

    @staticmethod
    def applyFrequencyFilter(originSeries, sampleFrequency=0, cutOffFrequency=0, filterType='low'):
        """Method applies low pass filter to series"""
        # Set cuttoff frequency if not given
        if sampleFrequency == 0:
            sampleFrequency = 1 / (originSeries.index[1] - originSeries.index[0])  # Sample distance
        if cutOffFrequency == 0:
            cutOffFrequency = sampleFrequency / 100  # Cut-off frequency of the filter
        normalizedFrequency = cutOffFrequency / (sampleFrequency / 2)  # Normalize the frequency

        b, a = s.butter(5, normalizedFrequency, filterType)
        filteredSeries = pd.DataFrame(data=s.filtfilt(b, a, originSeries.values), index=originSeries.index,
                                      columns=['filtered'])
        return filteredSeries['filtered']

    @staticmethod
    def normalizeDFWithGlobalMaxAndMin(df):
        df = df.dropna()
        return (df- np.min(df.values))/(np.max(df.values)-np.min(df.values))

    @staticmethod
    def normalizeDFofSignalGroup(df,signalNames):
        for signals in signalNames:
            df[signals] = SignalAnalysis.normalizeDFWithGlobalMaxAndMin(df[signals])
        return df
    
    @staticmethod
    def calculateSpectrogramFromSeries(originSeries):
        """Function calculates spectrogram from DF"""
        
        originSeries = originSeries.to_frame(name=None).set_index((originSeries.index-originSeries.index[0]))[originSeries.name]  #*10**-9
        ts = (originSeries.index[1]-originSeries.index[0]) #Sample distance
        T = (originSeries.index[-1]-originSeries.index[0]) #Timeperiod
        N = int(T/ts) # Number of samplepoints
        if len(originSeries) == N: #are the samples continuous
            newSeries = originSeries
        else:
            newSeries = SignalAnalysis.resampleSeries(originSeries)
        nperseg = N // 128
        try:
            f, t, Sxx = s.spectrogram(newSeries.values, 1/ts ,window=('hann'), nperseg=nperseg, noverlap = nperseg // 2)
            return f, t, Sxx
        except Exception as e:
            print('Could not calculate spectrogram: '+ str(e)) 

    @staticmethod
    def calculateFilterBankFromSpectrogram(spectrogram):
        """Function filterbank feature from sprectrogram"""
        if spectrogram is not None:
            f, t, Sxx = spectrogram
            filterBankMaxValues = np.zeros(len(Sxx[:,1])//8)
            for i in range(len(filterBankMaxValues)):
                k = i*8           
                filterBankMaxValues[i] = np.amax(Sxx[k:k+8,:])
            return np.mean(filterBankMaxValues[1:5])
        
    @staticmethod     
    def calculateCrossCorrelation(refSeries, measSeries):
        """Function compares to signals and returns the correlation coefficient, the length of the signals can vary"""
        shift = len(measSeries.values) - len(refSeries.values)
        maxCorrelationTimestamp = 0
        if shift > 0:
            R_val = np.zeros(abs(shift))
            i_val = np.zeros(abs(shift))

            for i in range(0 ,abs(shift)):
                i_val[i] += i
                R = np.corrcoef(measSeries.values[i:i+len(refSeries.values)], refSeries.values)
                R_val[i] = R[0][1]

            maxCorrelationTimestamp = measSeries.index[np.argmax(R_val)]
            maxCorrelation = np.max(R_val)

        elif shift < 0:
            R_val = np.zeros(abs(shift))
            i_val = np.zeros(abs(shift))

            for i in range(0 ,abs(shift)):
                i_val[i] += i
                R = np.corrcoef(refSeries.values[i:i+len(measSeries.values)], measSeries.values)
                R_val[i] = R[0][1]

            maxCorrelationTimestamp = refSeries.index[np.argmax(R_val)]
            maxCorrelation = np.max(R_val)
        else:
            R = np.corrcoef(measSeries.values, refSeries.values)
            maxCorrelation = R[0][1]
        return maxCorrelation,  maxCorrelationTimestamp

    @staticmethod
    def calculateSimilarityMatrix(myDF):
        simM = np.full(np.math.factorial(len(myDF.columns) - 1), -1)
        myVariable = np.nonzero(np.triu(squareform(pdist(myDF.T, metric='euclidean'))).flatten())[0]
        if len(myVariable) > 0:
            simM = np.triu(squareform(pdist(myDF.T, metric='euclidean'))).flatten()
        return pd.DataFrame(data=simM[np.nonzero(simM)], index=np.nonzero(simM)[0],
                            columns=[myDF.columns.values[0][:-2] + '_SimilarityMatrix'])

    @staticmethod
    def calculateCurve2CurveComparison(mySeriesDecay, mySeriesDecayRef, envelopeAmplitude,
                                       envelopeDecayTimeConstant=False, applyFilter=False, filterCutOffFrequency=100):
        end = np.min([len(mySeriesDecay), len(mySeriesDecayRef)])

        mySeriesDecay = mySeriesDecay.iloc[0:end]
        mySeriesDecayRef = mySeriesDecayRef.iloc[0:end]

        if applyFilter:
            mySeriesDecay = SignalAnalysis.applyFrequencyFilter(mySeriesDecay, filterCutOffFrequency,
                                                                filterType=applyFilter)
            mySeriesDecayRef = SignalAnalysis.applyFrequencyFilter(mySeriesDecayRef, filterCutOffFrequency,
                                                                   filterType=applyFilter)

        index = mySeriesDecay.index
        if envelopeDecayTimeConstant:
            envelope = np.exp(-index / envelopeDecayTimeConstant)
            substractArray = np.abs(mySeriesDecayRef.values - mySeriesDecay.values) / envelope
        else:
            substractArray = np.abs(mySeriesDecayRef.values - mySeriesDecay.values)

        outOfEnvelopeRatio = np.sum(substractArray > envelopeAmplitude) / len(substractArray)
        return outOfEnvelopeRatio

    @staticmethod
    def calculateDtwDistance(s1, s2):
        DTW = {}

        for i in range(len(s1)):
            DTW[(i, -1)] = float('inf')
        for i in range(len(s2)):
            DTW[(-1, i)] = float('inf')
        DTW[(-1, -1)] = 0

        for i in range(len(s1)):
            for j in range(len(s2)):
                dist = (s1[i] - s2[j]) ** 2
                DTW[(i, j)] = dist + min(DTW[(i - 1, j)], DTW[(i, j - 1)], DTW[(i - 1, j - 1)])

        return math.sqrt(DTW[len(s1) - 1, len(s2) - 1])
